@@ -23,7 +23,11 @@
  */
 
 #include "tn5250-private.h"
-#include "scs.h"
+
+/*
+#define DEBUG
+#define VERBOSE
+*/
 
 /* Define a default page size of 8.5x11 inches.  These numbers are a little
  * magic.  The iSeries sends a page size in terms of characters per inch,
@@ -48,8 +52,14 @@
 #define COURIER 1
 #define COURIER_BOLD 2
 
-int scs2pdf_process34 (int *curpos, int *boldchars);
+void scs2pdf_nl (Tn5250SCS * This);
+void scs2pdf_pp (Tn5250SCS * This);
+void scs2pdf_ff (Tn5250SCS * This);
 int scs2pdf_ahpp (int *curpos, int *boldchars);
+void scs2pdf_process2b (Tn5250SCS * This);
+void scs2pdf_default (Tn5250SCS * This);
+
+void do_newpage (Tn5250SCS * This);
 
 int pdf_header ();
 int pdf_catalog (int objnum, int outlinesobject, int pageobject);
@@ -65,23 +75,40 @@ int pdf_font (int objnum, int fontname);
 void pdf_xreftable (int objnum);
 void pdf_trailer (int offset, int size, int root);
 int pdf_process_char (char character, int flush);
+Tn5250SCS *tn5250_scs2pdf_new ();
+
+struct _Tn5250SCSPrivate
+{
+  int newfontsize;
+  unsigned long objcount;
+  unsigned long streamsize;
+  unsigned long filesize;
+  int fontsize;
+  unsigned int pagenumber;
+  int columncheck;
+  int boldchars;
+  int do_bold;
+  char text[255];
+  int newpage;
+};
 
-unsigned char curchar;
 unsigned char nextchar;
+Tn5250CharMap *map;
 
-struct _expanding_array {
-   int *data;
-   int elems;
-   int alloc;
+struct _expanding_array
+{
+  int *data;
+  int elems;
+  int alloc;
 };
 typedef struct _expanding_array expanding_array;
 
-expanding_array *array_new();
-void array_append_val(expanding_array *array, int value);
-int array_index(expanding_array *array, int idx);
-void array_free(expanding_array *array);
-   
+expanding_array *array_new ();
+void array_append_val (expanding_array * array, int value);
+int array_index (expanding_array * array, int idx);
+void array_free (expanding_array * array);
 
+expanding_array *textobjects;
 expanding_array *ObjectList;
 
 FILE *outfile;
@@ -89,23 +116,14 @@ FILE *outfile;
 int
 main ()
 {
-  Tn5250CharMap *map;
-  int pagewidth, pagelength;
-  int objcount = 0;
-  int filesize = 0;
-  int streamsize = 0;
-  int pagenumber;
-  int textobjects[10000];
   int pageparent, procsetobject, fontobject, boldfontobject, rootobject;
   int i;
-  int newpage = 0;
-  int column;
-  int new_line = 1;
-  int columncheck = 0;
-  int fontsize = 10, newfontsize;
-  int boldchars, do_bold;
-  char text[255];
+  Tn5250SCS *scs = NULL;
+
+
   ObjectList = array_new ();
+  textobjects = array_new ();
+
 
   /* This allows the user to select an output file other than stdout.
    * I don't know that this will ever be useful since you do pretty much
@@ -139,348 +157,346 @@ main ()
       map = tn5250_char_map_new ("37");
     }
 
+
+
+  /* Initialize the scs toolkit */
+  scs = tn5250_scs2pdf_new ();
+
+  if (scs == NULL)
+    {
+      return (-1);
+    }
+
+  scs->cpi = 0;
+  scs->data->newfontsize = 0;
+  scs->data->objcount = 0;
+  scs->data->streamsize = 0;
+  scs->data->filesize = 0;
+  scs->data->fontsize = 10;
+  scs->data->pagenumber = 1;
+  scs->column = 1;
+  scs->data->columncheck = 0;
+  scs->data->boldchars = 0;
+  scs->data->do_bold = 0;
+  scs->data->text[0] = '\0';
+  scs->data->newpage = 0;
+
+
   /* Write out the PDF header.  filesize tracks how big the PDF is since
    * we need that information later.
    */
-  filesize += pdf_header ();
+  scs->data->filesize += pdf_header ();
 
   /* ObjectList contains an entry for the filesize when the object was
    * created.  Since the cross reference of a PDF needs to know what the
    * byte count is for the beginning of each object we use ObjectList to 
    * track it.
    */
-  array_append_val (ObjectList, filesize);
-  objcount++;
-  filesize += pdf_begin_stream (objcount, COURIER, fontsize);
+  array_append_val (ObjectList, scs->data->filesize);
+  scs->data->objcount++;
+  scs->data->filesize +=
+    pdf_begin_stream (scs->data->objcount, COURIER, scs->data->fontsize);
 #ifdef DEBUG
-  fprintf (stderr, "objcount = %d\n", objcount);
+  fprintf (stderr, "objcount = %d\n", scs->data->objcount);
 #endif
 
-  pagenumber = 1;
-  column = 0;
-  boldchars = 0;
-  do_bold = 0;
-  newfontsize = 0;
 
-  while (!feof (stdin))
-    {
-      curchar = fgetc (stdin);
-      switch (curchar)
-	{
-	case SCS_TRANSPARENT:
-	  {
-	    scs_transparent ();
-	    break;
-	  }
-	case SCS_RFF:
-	  {
-	    scs_rff ();
-	    break;
-	  }
-	case SCS_NOOP:
-	  {
-	    scs_noop ();
-	    break;
-	  }
-	case SCS_CR:
-	  {
-	    scs_cr ();
-	    break;
-	  }
-	case SCS_FF:
-	  {
-	    scs_ff ();
-	    newpage = 1;
-	    break;
-	  }
-	case SCS_NL:
-	  {
-	    new_line = scs_nl ();
 
-	    /* If the current position on the page (column) is further to the
-	     * right than our current maximum line length (columncheck)
-	     * increase our length.  This is used later to see if we need a
-	     * larger page width if the page width is unspecified.
-	     */
-	    if (column > columncheck)
-	      {
-		columncheck = column;
-	      }
-	    column = 0;
-	    break;
-	  }
-	case SCS_RNL:
-	  {
-	    scs_rnl ();
-	    break;
-	  }
-	case SCS_HT:
-	  {
-	    scs_ht ();
-	    break;
-	  }
-	case 0x34:
-	  {
-	    boldchars = 0;
-	    streamsize += scs2pdf_process34 (&column, &boldchars);
+  /* Turn control over to the SCS toolkit and run the event loop */
+  scs_main (scs);
 
-	    /* If scs2pdf_process34() tells us that there are bold characters
-	     * to write and we aren't already in the middle of writing some
-	     * bold character, flush the buffer and send the bold font to the
-	     * PDF.
-	     */
-	    if ((boldchars > 0) && (do_bold == 0))
-	      {
-		do_bold = 1;
+
+  array_append_val (textobjects, scs->data->objcount);
+  scs->data->streamsize += pdf_process_char ('\0', 1);
+  scs->data->filesize += scs->data->streamsize;
+  scs->data->filesize += pdf_end_stream ();
+
+  array_append_val (ObjectList, scs->data->filesize);
+  scs->data->objcount++;
+  scs->data->filesize +=
+    pdf_stream_length (scs->data->objcount, scs->data->streamsize);
 #ifdef DEBUG
-		fprintf (stderr, "Starting bold font\n");
-#endif
-		streamsize += pdf_process_char ('\0', 1);
-		sprintf (text, "\t\t/F%d %d Tf\n", COURIER_BOLD, fontsize);
-		fprintf (outfile, "%s", text);
-		streamsize += strlen (text);
-	      }
-	    break;
-	  }
-	case 0x2B:
-	  {
-	    scs_process2b (&pagewidth, &pagelength, &newfontsize);
-
-	    /* If scs_process2b() wants to change the CPI (newfontsize), flush
-	     * the buffer and send the new font to the PDF.  Note that right
-	     * now we always reset the font to COURIER, even if we are using
-	     * a different font.  This will be changed soon.
-	     */
-	    if (newfontsize > 0)
-	      {
-#ifdef DEBUG
-		fprintf (stderr, "Changing font size to %d\n", newfontsize);
-#endif
-		streamsize += pdf_process_char ('\0', 1);
-		sprintf (text, "\t\t/F%d %d Tf\n", COURIER, newfontsize);
-		fprintf (outfile, "%s", text);
-		streamsize += strlen (text);
-		fontsize = newfontsize;
-		newfontsize = 0;
-	      }
-	    break;
-	  }
-	case 0xFF:
-	  {
-	    /* This is a hack */
-	    /* Don't know where the 0xFF is coming from */
-	    break;
-	  }
-	default:
-	  {
-	    if (newpage)
-	      {
-		/* We need to know what stream objects (textobjects) were
-		 * put on this page.  We put one stream object on each
-		 * page.
-		 */
-		textobjects[pagenumber - 1] = objcount;
-
-		streamsize += pdf_process_char ('\0', 1);
-		filesize += streamsize;
-		filesize += pdf_end_stream ();
-
-		array_append_val (ObjectList, filesize);
-		objcount++;
-		filesize += pdf_stream_length (objcount, streamsize);
-		streamsize = 0;
-#ifdef DEBUG
-		fprintf (stderr, "objcount = %d\n", objcount);
+  fprintf (stderr, "stream length objcount = %d\n", scs->data->objcount);
 #endif
 
-		/* Ending the stream object above and starting a new one
-		 * here constitutes a new page, in conjuction with the
-		 * pdf_page() function below.
-		 */
-		array_append_val (ObjectList, filesize);
-		objcount++;
-		filesize += pdf_begin_stream (objcount, COURIER, fontsize);
+  array_append_val (ObjectList, scs->data->filesize);
+  scs->data->objcount++;
+  scs->data->filesize += pdf_catalog (scs->data->objcount,
+				      scs->data->objcount + 1,
+				      scs->data->objcount + 5);
+  rootobject = scs->data->objcount;
 #ifdef DEBUG
-		fprintf (stderr, "objcount = %d\n", objcount);
+  fprintf (stderr, "catalog objcount = %d\n", scs->data->objcount);
 #endif
 
-		pagenumber++;
-		newpage = 0;
-		column = 0;
-	      }
-	    if (new_line)
-	      {
-		/* On newline flush the buffer and move the active line down
-		 * 12 points.
-		 */
-		new_line = 0;
-		streamsize += pdf_process_char ('\0', 1);
-		fprintf (outfile, "0 -12 Td\n");
-		streamsize += 9;
-		column = 0;
-	      }
-	    streamsize += pdf_process_char (tn5250_char_map_to_local (map,
-								      curchar),
-					    0);
-	    /* If you want to feed this program non-EBCDIC text then uncomment
-	     * the line below and comment the line above.
-	     */
-	    /*
-	     * streamsize += pdf_process_char(curchar, 0);
-	     */
-	    column++;
-
-	    /* If we are currently printing bold character then decrement the
-	     * bold character count (boldchars) given above by
-	     * scs2pdf_process34() until there aren't any bold characters left
-	     * to print.  When out of bold character reset the font and allow
-	     * changing to bold again.
-	     */
-	    if (boldchars > 0)
-	      {
-		boldchars--;
-		if (boldchars == 0)
-		  {
+  array_append_val (ObjectList, scs->data->filesize);
+  scs->data->objcount++;
+  scs->data->filesize += pdf_outlines (scs->data->objcount);
 #ifdef DEBUG
-		    fprintf (stderr, "Ending bold font\n");
-#endif
-		    do_bold = 0;
-		    streamsize += pdf_process_char ('\0', 1);
-		    sprintf (text, "\t\t/F%d %d Tf\n", COURIER, fontsize);
-		    fprintf (outfile, "%s", text);
-		    streamsize += strlen (text);
-		  }
-	      }
-	  }
-	}
-
-    }
-  textobjects[pagenumber - 1] = objcount;
-  streamsize += pdf_process_char ('\0', 1);
-  filesize += streamsize;
-  filesize += pdf_end_stream ();
-
-  array_append_val (ObjectList, filesize);
-  objcount++;
-  filesize += pdf_stream_length (objcount, streamsize);
-#ifdef DEBUG
-  fprintf (stderr, "stream length objcount = %d\n", objcount);
+  fprintf (stderr, "outlines objcount = %d\n", scs->data->objcount);
 #endif
 
-  array_append_val (ObjectList, filesize);
-  objcount++;
-  filesize += pdf_catalog (objcount, objcount + 1, objcount + 5);
-  rootobject = objcount;
+  array_append_val (ObjectList, scs->data->filesize);
+  scs->data->objcount++;
+  scs->data->filesize += pdf_procset (scs->data->objcount);
+  procsetobject = scs->data->objcount;
 #ifdef DEBUG
-  fprintf (stderr, "catalog objcount = %d\n", objcount);
-#endif
-
-  array_append_val (ObjectList, filesize);
-  objcount++;
-  filesize += pdf_outlines (objcount);
-#ifdef DEBUG
-  fprintf (stderr, "outlines objcount = %d\n", objcount);
-#endif
-
-  array_append_val (ObjectList, filesize);
-  objcount++;
-  filesize += pdf_procset (objcount);
-  procsetobject = objcount;
-#ifdef DEBUG
-  fprintf (stderr, "procedure set objcount = %d\n", objcount);
+  fprintf (stderr, "procedure set objcount = %d\n", scs->data->objcount);
 #endif
 
   /* We need to create a font object for every font we use.  Since we don't
    * necessarily know if we used bold just make a bold font object anyway.
    * It doesn't hurt to have objects that aren't used.
    */
-  array_append_val (ObjectList, filesize);
-  objcount++;
-  filesize += pdf_font (objcount, COURIER);
-  fontobject = objcount;
+  array_append_val (ObjectList, scs->data->filesize);
+  scs->data->objcount++;
+  scs->data->filesize += pdf_font (scs->data->objcount, COURIER);
+  fontobject = scs->data->objcount;
 #ifdef DEBUG
-  fprintf (stderr, "font objcount = %d\n", objcount);
+  fprintf (stderr, "font objcount = %d\n", scs->data->objcount);
 #endif
 
-  array_append_val (ObjectList, filesize);
-  objcount++;
-  filesize += pdf_font (objcount, COURIER_BOLD);
-  boldfontobject = objcount;
+  array_append_val (ObjectList, scs->data->filesize);
+  scs->data->objcount++;
+  scs->data->filesize += pdf_font (scs->data->objcount, COURIER_BOLD);
+  boldfontobject = scs->data->objcount;
 #ifdef DEBUG
-  fprintf (stderr, "bold font objcount = %d\n", objcount);
+  fprintf (stderr, "bold font objcount = %d\n", scs->data->objcount);
 #endif
 
   /* If the page size was not specified and we have lines longer than we
    * think we can fit on a page increase the page width a half inch at a time.
    */
-  if ((pagewidth == 0) && (columncheck > MAX_COLUMNS))
+  if ((scs->pagewidth == 0) && (scs->data->columncheck > MAX_COLUMNS))
     {
       for (i = 0; i < 15; i++)
 	{
-	  if (columncheck < (MAX_COLUMNS + (5 * i)))
+	  if (scs->data->columncheck < (MAX_COLUMNS + (5 * i)))
 	    {
 #ifdef DEBUG
 	      fprintf (stderr, "columncheck = %d, pagewidth = %d\n",
-		       columncheck, pagewidth);
+		       scs->data->columncheck, scs->pagewidth);
 #endif
 	      break;
 	    }
-	  pagewidth = DEFAULT_PAGE_WIDTH + (720 * i);
+	  scs->pagewidth = DEFAULT_PAGE_WIDTH + (720 * i);
 	}
     }
-  array_append_val (ObjectList, filesize);
-  objcount++;
-  filesize += pdf_pages (objcount, objcount + 1, pagenumber);
-  pageparent = objcount;
+  array_append_val (ObjectList, scs->data->filesize);
+  scs->data->objcount++;
+  scs->data->filesize += pdf_pages (scs->data->objcount,
+				    scs->data->objcount + 1,
+				    scs->data->pagenumber);
+  pageparent = scs->data->objcount;
 #ifdef DEBUG
-  fprintf (stderr, "pages objcount = %d\n", objcount);
+  fprintf (stderr, "pages objcount = %d\n", scs->data->objcount);
 #endif
 
-  for (i = 0; i < pagenumber; i++)
+  for (i = 0; i < scs->data->pagenumber; i++)
     {
-      array_append_val (ObjectList, filesize);
-      objcount++;
-      filesize += pdf_page (objcount,
-			    pageparent,
-			    textobjects[i], procsetobject, fontobject,
-			    boldfontobject, pagewidth, pagelength);
+      array_append_val (ObjectList, scs->data->filesize);
+      scs->data->objcount++;
+      scs->data->filesize += pdf_page (scs->data->objcount,
+				       pageparent,
+				       array_index (textobjects, i),
+				       procsetobject, fontobject,
+				       boldfontobject, scs->pagewidth,
+				       scs->pagelength);
 #ifdef DEBUG
-      fprintf (stderr, "page objcount = %d\n", objcount);
+      fprintf (stderr, "page objcount = %d\n", scs->data->objcount);
 #endif
     }
 
-  pdf_xreftable (objcount);
-  pdf_trailer (filesize, objcount + 1, rootobject);
+  pdf_xreftable (scs->data->objcount);
+  pdf_trailer (scs->data->filesize, scs->data->objcount + 1, rootobject);
 
+  array_free (textobjects);
   array_free (ObjectList);
   tn5250_char_map_destroy (map);
+  free (scs);
   return (0);
 }
 
 
 
+/* This initializes the scs callbacks
+ */
+Tn5250SCS *
+tn5250_scs2pdf_new ()
+{
+  Tn5250SCS *scs = tn5250_scs_new ();
+
+  if (scs == NULL)
+    {
+      fprintf (stderr,
+	       "Unable to allocate memory in tn5250_scs2pdf_new ()!\n");
+      return NULL;
+    }
+
+  scs->data = tn5250_new (struct _Tn5250SCSPrivate, 1);
+  if (scs->data == NULL)
+    {
+      free (scs);
+      return NULL;
+    }
+
+  /* And now set up our callbacks */
+  scs->pp = scs2pdf_pp;
+  scs->nl = scs2pdf_nl;
+  scs->rnl = scs2pdf_nl;
+  scs->ff = scs2pdf_ff;
+  scs->rff = scs2pdf_ff;
+  scs->process2b = scs2pdf_process2b;
+  scs->scsdefault = scs2pdf_default;
+  return scs;
+}
+
+
+void
+scs2pdf_default (Tn5250SCS * This)
+{
+  if (This->data->newpage == 1)
+    {
+      do_newpage (This);
+    }
+
+  This->data->streamsize +=
+    pdf_process_char (tn5250_char_map_to_local (map, This->curchar), 0);
+
+  /* If you want to feed this program non-EBCDIC text then uncomment
+   * the line below and comment the line above.
+   *
+   * streamsize += pdf_process_char(curchar, 0);
+   */
+  This->column = This->column + 1;
+
+  /* If we are currently printing bold character then decrement the
+   * bold character count (boldchars) given above by
+   * scs2pdf_pp() until there aren't any bold characters left
+   * to print.  When out of bold character reset the font and allow
+   * changing to bold again.
+   */
+  if (This->data->boldchars > 0)
+    {
+      This->data->boldchars = This->data->boldchars - 1;
+      if (This->data->boldchars == 0)
+	{
+#ifdef DEBUG
+	  fprintf (stderr, "Ending bold font\n");
+#endif
+	  This->data->do_bold = 0;
+	  This->data->streamsize += pdf_process_char ('\0', 1);
+	  sprintf (This->data->text, "\t\t/F%d %d Tf\n",
+		   COURIER, This->data->fontsize);
+	  fprintf (outfile, "%s", This->data->text);
+	  This->data->streamsize += strlen (This->data->text);
+	}
+    }
+  return;
+}
+
+
+
+/* Custom process 0x2B function */
+void
+scs2pdf_process2b (Tn5250SCS * This)
+{
+  scs_process2b (This);
+
+  /* If scs_process2b() wants to change the CPI (newfontsize), flush
+   * the buffer and send the new font to the PDF.  Note that right
+   * now we always reset the font to COURIER, even if we are using
+   * a different font.  This will be changed soon.
+   */
+  if (This->cpi > 0)
+    {
+#ifdef DEBUG
+      fprintf (stderr, "Changing font size to %d\n", This->cpi);
+#endif
+      This->data->streamsize += pdf_process_char ('\0', 1);
+      sprintf (This->data->text, "\t\t/F%d %d Tf\n", COURIER, This->cpi);
+      fprintf (outfile, "%s", This->data->text);
+      This->data->streamsize += strlen (This->data->text);
+      This->data->fontsize = This->cpi;
+      This->cpi = 0;
+    }
+  return;
+}
+
+
+
+/* Handle new lines uniquely to track column number
+ */
+void
+scs2pdf_nl (Tn5250SCS * This)
+{
+  scs_nl (This);
+
+  /* If the current position on the page (column) is further to the
+   * right than our current maximum line length (columncheck)
+   * increase our length.  This is used later to see if we need a
+   * larger page width if the page width is unspecified.
+   */
+  if (This->column > This->data->columncheck)
+    {
+      This->data->columncheck = This->column;
+    }
+  /* On newline flush the buffer and move the active line down
+   * 12 points.
+   */
+  This->data->streamsize += pdf_process_char ('\0', 1);
+  fprintf (outfile, "0 -12 Td\n");
+  This->data->streamsize += 9;
+  This->column = 1;
+  return;
+}
+
+
+
+void
+scs2pdf_ff (Tn5250SCS * This)
+{
+  scs_ff (This);
+  This->data->newpage = 1;
+  return;
+}
+
+
+
 /* This function is different than what is in scs.c because we want to pass
  * a pointer to the number of bold characters to print to scs2pdf_ahpp()
  * (which is a unique version of scs_ahpp()).  Since this program is the only
  * one that handles bold this is unique.
  */
-int
-scs2pdf_process34 (int *curpos, int *boldchars)
+void
+scs2pdf_pp (Tn5250SCS * This)
 {
+  unsigned char curchar;
   int bytes;
 
   bytes = 0;
+  This->data->boldchars = 0;
   curchar = fgetc (stdin);
   switch (curchar)
     {
+    case SCS_RDPP:
+      {
+	scs_rdpp (NULL);
+	break;
+      }
     case SCS_AVPP:
       {
-	scs_avpp ();
+	scs_avpp (NULL);
 	break;
       }
     case SCS_AHPP:
       {
-	bytes = scs2pdf_ahpp (curpos, boldchars);
+	bytes = scs2pdf_ahpp (&(This->column), &(This->data->boldchars));
+	break;
+      }
+    case SCS_RRPP:
+      {
+	scs_rrpp (NULL);
 	break;
       }
     default:
@@ -488,7 +504,26 @@ scs2pdf_process34 (int *curpos, int *boldchars)
 	fprintf (stderr, "ERROR: Unknown 0x34 command %x\n", curchar);
       }
     }
-  return (bytes);
+  This->data->streamsize += bytes;
+
+  /* If scs2pdf_ahpp() tells us that there are bold characters
+   * to write and we aren't already in the middle of writing some
+   * bold character, flush the buffer and send the bold font to the
+   * PDF.
+   */
+  if ((This->data->boldchars > 0) && (This->data->do_bold == 0))
+    {
+      This->data->do_bold = 1;
+#ifdef DEBUG
+      fprintf (stderr, "Starting bold font\n");
+#endif
+      This->data->streamsize += pdf_process_char ('\0', 1);
+      sprintf (This->data->text, "\t\t/F%d %d Tf\n",
+	       COURIER_BOLD, This->data->fontsize);
+      fprintf (outfile, "%s", This->data->text);
+      This->data->streamsize += strlen (This->data->text);
+    }
+  return;
 }
 
 
@@ -505,8 +540,11 @@ scs2pdf_ahpp (int *curpos, int *boldchars)
 
   bytes = 0;
   position = fgetc (stdin);
+#ifdef DEBUG
+  fprintf (stderr, "AHPP %d (current position: %d)\n", position, *curpos);
+#endif
 
-  if (*curpos > position)
+  if ((*curpos - 1) > position)
     {
       /* Frank Richter <frichter@esda.com> noticed that we should be
        * going back and printing over the same line if *curpos is greater
@@ -523,7 +561,7 @@ scs2pdf_ahpp (int *curpos, int *boldchars)
       fprintf (outfile, "0 0 Td\n");
       bytes += 7;
 
-      for (i = 0; i < position; i++)
+      for (i = 0; i < position - 1; i++)
 	{
 	  bytes += pdf_process_char (' ', 0);
 	}
@@ -537,14 +575,59 @@ scs2pdf_ahpp (int *curpos, int *boldchars)
     }
   *curpos = position;
 
-#ifdef DEBUG
-  fprintf (stderr, "AHPP %d\n", position);
-#endif
   return (bytes);
 }
 
 
 
+void
+do_newpage (Tn5250SCS * This)
+{
+  /* We need to know what stream objects (textobjects) were
+   * put on this page.  We put one stream object on each
+   * page.
+   */
+  array_append_val (textobjects, This->data->objcount);
+
+  This->data->streamsize += pdf_process_char ('\0', 1);
+  This->data->filesize += This->data->streamsize;
+  This->data->filesize += pdf_end_stream ();
+
+  array_append_val (ObjectList, This->data->filesize);
+  This->data->objcount = This->data->objcount + 1;
+#ifdef DEBUG
+  fprintf (stderr, "objcount: %d\n", This->data->objcount);
+#endif
+  This->data->filesize += pdf_stream_length (This->data->objcount,
+					     This->data->streamsize);
+  This->data->streamsize = 0;
+#ifdef DEBUG
+  fprintf (stderr, "objcount = %d\n", This->data->objcount);
+#endif
+
+  /* Ending the stream object above and starting a new one
+   * here constitutes a new page, in conjuction with the
+   * pdf_page() function below.
+   */
+  array_append_val (ObjectList, This->data->filesize);
+  This->data->objcount = This->data->objcount + 1;
+  This->data->filesize += pdf_begin_stream (This->data->objcount,
+					    COURIER, This->data->fontsize);
+#ifdef DEBUG
+  fprintf (stderr, "objcount = %d\n", This->data->objcount);
+#endif
+
+  This->data->pagenumber++;
+#ifdef DEBUG
+  fprintf (stderr, "pagenumber: %d\n", This->data->pagenumber);
+#endif
+  This->column = 1;
+  This->data->newpage = 0;
+  return;
+}
+
+
+
 /* This header is required on all PDFs to identify what level of the PDF
  * specification was used to create this PDF.
  */
@@ -854,8 +937,7 @@ pdf_xreftable (int objnum)
    */
   for (curobj = 0; curobj < objnum; curobj++)
     {
-      fprintf (outfile,
-	       "%010d 00000 n \n", array_index (ObjectList, curobj));
+      fprintf (outfile, "%010d 00000 n \n", array_index (ObjectList, curobj));
     }
 }
 
@@ -926,58 +1008,71 @@ pdf_process_char (char character, int flush)
 }
 
 
-expanding_array *array_new() {
+expanding_array *
+array_new ()
+{
+  expanding_array *array = NULL;
 
-   expanding_array *array = NULL;
+  array = malloc (sizeof (expanding_array));
+  if (array == NULL)
+    return NULL;
 
-   array = malloc(sizeof(expanding_array));
-   if (array == NULL) 
-        return NULL;
+  array->data = NULL;
+  array->elems = 0;
+  array->alloc = 0;
 
-   array->data = NULL;
-   array->elems = 0;
-   array->alloc = 0;
-   
-   return array;
+  return array;
 }
 
 
-void array_append_val(expanding_array *array, int value) {
+void
+array_append_val (expanding_array * array, int value)
+{
+  int *newdata;
 
-   int *newdata;
+  array->elems++;
 
-   array->elems ++;
-
-   if (array->elems > array->alloc) {
-       array->alloc += 64;
-       if (array->data == NULL) {
-            array->data = malloc(sizeof(int) * array->alloc);
-       } else {
-            newdata = realloc(array->data, sizeof(int) * array->alloc);
-            if (newdata == NULL) 
-                free(array->data);
-            array->data = newdata;
-       }
-       if (array->data == NULL) {
-            fprintf(stderr, "array_append_val: Out of memory!\n");
-            exit(1);
-       }
-   }
-        
-   array->data[array->elems - 1] = value;
+  if (array->elems > array->alloc)
+    {
+      array->alloc += 64;
+      if (array->data == NULL)
+	{
+	  array->data = malloc (sizeof (int) * array->alloc);
+	}
+      else
+	{
+	  newdata = realloc (array->data, sizeof (int) * array->alloc);
+	  if (newdata == NULL)
+	    free (array->data);
+	  array->data = newdata;
+	}
+      if (array->data == NULL)
+	{
+	  fprintf (stderr, "array_append_val: Out of memory!\n");
+	  exit (1);
+	}
+    }
 
+  array->data[array->elems - 1] = value;
+  return;
 }
 
-int array_index(expanding_array *array, int idx) {
 
-    return array->data[idx];
+int
+array_index (expanding_array * array, int idx)
+{
+  return array->data[idx];
+}
 
+
+void
+array_free (expanding_array * array)
+{
+  if (array->data != NULL)
+    {
+      free (array->data);
+    }
+  free (array);
+  array = NULL;
+  return;
 }
-     
-void array_free(expanding_array *array) {
-    if (array->data != NULL) 
-        free(array->data);
-    free(array);
-    array = NULL;
-}
-/* vi:set sts=3 sw=3: */
