@@ -88,6 +88,7 @@ static void slang_terminal_update_indicators(Tn5250Terminal * This,
 static int slang_terminal_waitevent(Tn5250Terminal * This);
 static int slang_terminal_getkey(Tn5250Terminal * This);
 static void slang_terminal_beep(Tn5250Terminal * This);
+static int slang_terminal_enhanced (Tn5250Terminal * This);
 static int slang_terminal_get_esc_key(Tn5250Terminal * This, int is_esc);
 static void slang_terminal_set_attrs (Tn5250Terminal * This, int attrs);
 
@@ -135,7 +136,9 @@ Tn5250Terminal *tn5250_slang_terminal_new()
    r->update_indicators = slang_terminal_update_indicators;
    r->waitevent = slang_terminal_waitevent;
    r->getkey = slang_terminal_getkey;
+   r->putkey = NULL;
    r->beep = slang_terminal_beep;
+   r->flags = slang_terminal_enhanced;
    return r;
 }
 
@@ -370,6 +373,9 @@ static void slang_terminal_update_indicators(Tn5250Terminal * This, Tn5250Displa
    if ((inds & TN5250_DISPLAY_IND_INSERT) != 0) {
       memcpy(ind_buf + 30, "IM", 2);
    }
+   if ((inds & TN5250_DISPLAY_IND_MACRO) != 0) {
+      memcpy(ind_buf + 54, tn5250_macro_printstate (display), 11);
+   }
    SLsmg_normal_video ();
    SLsmg_gotorc (tn5250_display_height (display), 0);
    SLsmg_write_nchars (ind_buf, 80);
@@ -472,10 +478,14 @@ static int slang_terminal_getkey(Tn5250Terminal * This)
 	 return K_PRINT;
       case K_CTRL('R'):
 	 return K_RESET;	/* Error Reset */
+      case K_CTRL('S'):
+	 return K_MEMO;
       case K_CTRL('T'):
 	 return K_TESTREQ;
       case K_CTRL('U'):
 	 return K_ROLLDN;
+      case K_CTRL('W'):
+	 return K_EXEC;
       case K_CTRL('X'):
 	 return K_FIELDEXIT;
 
@@ -601,6 +611,24 @@ static void slang_terminal_beep (Tn5250Terminal * This)
    SLtt_beep ();
 }
 
+
+/***** lib5250/slang_terminal_enhanced
+ * NAME
+ *    slang_terminal_enhanced
+ * SYNOPSIS
+ *    ret = slang_terminal_enhanced (This);
+ * INPUTS
+ *    Tn5250Terminal  *    This       - 
+ * DESCRIPTION
+ *    Return 1 if we support the enhanced 5250 protocol, 0 otherwise.
+ *****/
+static int
+slang_terminal_enhanced (Tn5250Terminal * This)
+{
+  return (0);
+}
+
+
 /****i* lib5250/slang_terminal_get_esc_key
  * NAME
  *    slang_terminal_get_esc_key
