@@ -33,7 +33,6 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
-#include <malloc.h>
 #include <ctype.h>
 #include <errno.h>
 #include <signal.h>
@@ -77,12 +76,16 @@
 #include "dbuffer.h"
 #include "field.h"
 #include "codes5250.h"
-#include "terminal.h"
+#include "scrollbar.h"
 #include "session.h"
 #include "printsession.h"
 #include "display.h"
-#include "debug.h"
+#include "macro.h"
+#include "menu.h"
 #include "wtd.h"
+#include "window.h"
+#include "terminal.h"
+#include "debug.h"
 #include "scs.h"
 #include "conf.h"
 
