@@ -1,4 +1,4 @@
 
-#include "config.h"
+#include "tn5250-config.h"
 
 char *version_string = VERSION;
