@@ -32,14 +32,17 @@ char *version_string = "not implemented";
 
 #define LP5250D_WINDOW_CLASS "lp5250d_window_class"
 #define WM_TN5250_STREAM_DATA WM_USER+2000
+#define WM_TN5250_NOTIFY_ICON WM_USER+2001
 
 WNDCLASSEX wndclass;
 HWND hWnd;
+int WindowIsVisible;
 
 Tn5250PrintSession *printsess = NULL;
 Tn5250Stream *stream = NULL;
 Tn5250Config *config = NULL;
 char globStatus[MAXMSG+1];
+NOTIFYICONDATA nid;
 
 struct _Tn5250Printer {
      int openpg;
@@ -133,8 +136,9 @@ int WINAPI WinMain(HINSTANCE i, HINSTANCE p, PSTR cmd, int show)
    GetClientRect(hWnd, &rec);
    MoveWindow(hWnd, rec.left, rec.top, newwidth, newheight, FALSE);
 
-   ShowWindow(hWnd, show);
-   UpdateWindow(hWnd);
+   ShowWindow(hWnd, SW_HIDE);
+   ShowWindow(hWnd, SW_HIDE);
+   WindowIsVisible = 0;
 
    /* Start up Windows Sockets, and make sure we've got a compatable
       winsock layer */
@@ -150,6 +154,16 @@ int WINAPI WinMain(HINSTANCE i, HINSTANCE p, PSTR cmd, int show)
 	exit(1);
     }
 
+   /* make us appear in system tray */
+   nid.cbSize = sizeof(nid);
+   nid.hWnd = hWnd;
+   nid.uID =  1;
+   nid.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
+   nid.uCallbackMessage = WM_TN5250_NOTIFY_ICON;
+   nid.hIcon = LoadIcon(i, MAKEINTRESOURCE(IDI_TN5250_ICON));
+   strcpy(nid.szTip, "5250 printer");
+   Shell_NotifyIcon(NIM_ADD, &nid);
+
    /* parse the "cmd" variable into a seperate string for each argument */
    argc = parse_cmdline(cmd, NULL);
    if (argc>0) {
@@ -163,6 +177,7 @@ int WINAPI WinMain(HINSTANCE i, HINSTANCE p, PSTR cmd, int show)
    config = tn5250_config_new ();
    if (tn5250_config_load_default (config) == -1) {
         tn5250_config_unref(config);
+        Shell_NotifyIcon(NIM_DELETE, &nid);
         exit (1);  
    }
    if (tn5250_config_parse_argv (config, argc, argv) == -1) {
@@ -177,6 +192,7 @@ int WINAPI WinMain(HINSTANCE i, HINSTANCE p, PSTR cmd, int show)
        syntax();
    else if (tn5250_config_get (config, "version")) {
        msgboxf ("tn5250 version %s\n", version_string);
+       Shell_NotifyIcon(NIM_DELETE, &nid);
        exit (0);
    }
    else if (!tn5250_config_get (config, "host"))   
@@ -200,6 +216,7 @@ int WINAPI WinMain(HINSTANCE i, HINSTANCE p, PSTR cmd, int show)
     if (stream == NULL) {
        msgboxf("Couldn't connect to %s", 
 		         tn5250_config_get (config,"host"));
+       Shell_NotifyIcon(NIM_DELETE, &nid);
        exit(1);
     }
        
@@ -255,6 +272,7 @@ static void syntax()
 	  "\toutputcommand=CMD		specify the print output command\n"
 	  "\t+/-version			display version\n");
 
+   Shell_NotifyIcon(NIM_DELETE, &nid);
    exit (255);
 }
 
@@ -359,7 +377,6 @@ Tn5250Printer * tn5250_windows_printer_startdoc(void) {
     char *temp;
 
     devname = malloc(MAXDEVNAME+1);
-    TN5250_ASSERT(devname!=NULL);
 
     if (tn5250_config_get(config, "outputcommand")) {
         len = strlen(tn5250_config_get(config, "outputcommand"));
@@ -386,7 +403,6 @@ Tn5250Printer * tn5250_windows_printer_startdoc(void) {
     }
 
     This = malloc(sizeof(Tn5250Printer));
-    TN5250_ASSERT(This!=NULL);
 
     if (tn5250_config_get (config, "map")) 
         This->map = tn5250_char_map_new (tn5250_config_get (config, "map"));
@@ -1041,6 +1057,7 @@ void tn5250_windows_print_session_main_loop(Tn5250PrintSession * This)
    int newjob;
    char responsecode[5];
    void *prthdl;
+   StreamHeader header;
 
 
    while (1) {
@@ -1053,13 +1070,16 @@ void tn5250_windows_print_session_main_loop(Tn5250PrintSession * This)
 	       if (This->rec != NULL)
 	          tn5250_record_destroy(This->rec);
 	       This->rec = tn5250_stream_get_record(This->stream);
-	       if (!tn5250_windows_print_session_get_response_code(This, responsecode))
+	       if (!tn5250_windows_print_session_get_response_code(This, responsecode)) {
+                  Shell_NotifyIcon(NIM_DELETE, &nid);
 	          exit (1);
+               }
 	       break;
 	    }
 	 }
 	 else {
             tn5250_windows_printer_status(2,"Socket closed by host.\n");
+            Shell_NotifyIcon(NIM_DELETE, &nid);
 	    exit(-1);
 	 }
       }
@@ -1089,11 +1109,14 @@ void tn5250_windows_print_session_main_loop(Tn5250PrintSession * This)
                   continue;
                }
 
+               header.h5250.flowtype = TN5250_RECORD_FLOW_CLIENTO;
+               header.h5250.flags = TN5250_RECORD_H_NONE;
+               header.h5250.opcode = TN5250_RECORD_OPCODE_PRINT_COMPLETE;
+
 	       tn5250_stream_send_packet(This->stream, 0,
-		     TN5250_RECORD_FLOW_CLIENTO,
-		     TN5250_RECORD_H_NONE,
-		     TN5250_RECORD_OPCODE_PRINT_COMPLETE,
+                     header,
 		     NULL);
+
 	       if (tn5250_record_length(This->rec) == 0x11) {
                   tn5250_windows_printer_status(0,"Job Complete\n");
                   tn5250_windows_printer_finishdoc(prthdl);
@@ -1107,6 +1130,7 @@ void tn5250_windows_print_session_main_loop(Tn5250PrintSession * This)
 	 }
 	 else {
             tn5250_windows_printer_status(2,"Socket closed by host\n");
+            Shell_NotifyIcon(NIM_DELETE, &nid);
 	    exit(-1);
 	 }
       }
@@ -1199,6 +1223,12 @@ LRESULT CALLBACK tn5250_windows_printer_wndproc(
            break;
        }
 
+       case WM_WINDOWPOSCHANGING:
+          break;
+
+       case WM_CLOSE:
+          Shell_NotifyIcon(NIM_DELETE, &nid);
+          break;
 
        case WM_DESTROY:
           PostQuitMessage(0);
@@ -1224,17 +1254,45 @@ LRESULT CALLBACK tn5250_windows_printer_wndproc(
           return 0;
           break;
 
+       case WM_SYSCOMMAND:
+          switch (wParam) {
+             case SC_RESTORE:
+                WindowIsVisible = 1;
+                ShowWindow(hwnd, SW_SHOW);
+                break;
+             case SC_MINIMIZE:
+                WindowIsVisible = 0;
+                ShowWindow(hwnd, SW_HIDE);
+                return 0;
+          }
+          break;
+                 
 
        case WM_COMMAND:
           if (LOWORD(wParam) == 1 &&
               HIWORD(wParam) == BN_CLICKED &&
               (HWND) lParam == hwndStopButton) 
           {
+              Shell_NotifyIcon(NIM_DELETE, &nid);
               DestroyWindow (hwnd);
           }
           return 0;
           break;
 
+      case WM_TN5250_NOTIFY_ICON:
+          switch (lParam) {
+             case WM_LBUTTONDBLCLK:
+               if (WindowIsVisible) {
+                  WindowIsVisible = 0;
+                  ShowWindow(hwnd, SW_HIDE);
+               }
+               else {
+                  WindowIsVisible = 1;
+                  ShowWindow(hwnd, SW_SHOW);
+               }  
+             break;
+          }
+          return 0;
 
       case WM_TN5250_STREAM_DATA:
           /* this is our special message to tell us that network data
@@ -1269,7 +1327,6 @@ int parse_cmdline(char *cmdline, char **my_argv) {
     
     if (my_argv!=NULL) {
          my_argv[0] = malloc(MAXARG+1);
-         TN5250_ASSERT(my_argv[0]!=NULL);
          GetModuleFileName(NULL, my_argv[0], MAXARG);
     }
 
@@ -1291,7 +1348,6 @@ int parse_cmdline(char *cmdline, char **my_argv) {
                        if (my_argv!=NULL) {
                             my_argv[argcnt] = 
                                    (char *)malloc((arglen+1) * sizeof(char));
-                            TN5250_ASSERT(my_argv[argcnt]!=NULL);
                             strcpy(my_argv[argcnt], arg);
                        }
                        argcnt++;
@@ -1325,7 +1381,6 @@ int parse_cmdline(char *cmdline, char **my_argv) {
     if (arglen>0 && argcnt<MAXARG) {
         if (my_argv!=NULL) {
             my_argv[argcnt] = (char *)malloc((arglen+1) * sizeof(char));
-            TN5250_ASSERT(my_argv[argcnt]!=NULL);
             strcpy(my_argv[argcnt], arg);
         }
         argcnt++;
