@@ -70,6 +70,7 @@ replace_undef redefs[] =
 #define DEF_PACKAGE        6
 #define DEF_VERSION        7
 #define DEF_BINARY         8
+#define DEF_NETSHARE400    9
 
 struct _manual_def {
      char *from;
@@ -87,6 +88,7 @@ manual_def mredefs[] =
   {"PACKAGE",        DEF_PACKAGE},
   {"VERSION",        DEF_VERSION},
   {"BINARY_RELEASE", DEF_BINARY},
+  {"NETSHARE400",    DEF_NETSHARE400},
   {NULL, -1},
 };
 
@@ -98,8 +100,8 @@ void replaceall(char *str, const char *from, const char *to);
 void strtrim(char *str);
 void get_path(const char *prompt, const char *file, char *path, int maxlen);
 int create_tn5250_config_h(const char *infile, const char *outfile, 
-                   char with_openssl, char binary_release,
-                   const char *package, const char *version);
+                   char with_openssl, char binary_release, 
+                   const char *package, const char *version, char netshare400);
 int create_config_h(const char *infile, const char *outfile, 
                         const char *version);
 int create_makefile(const char *infile, const char *outfile,
@@ -111,7 +113,7 @@ int create_makefile(const char *infile, const char *outfile,
 void replacedata(const char *from, const char *to, char *line, int maxlen);
 int make_root_makefile(const char *fn);
 char *manual_redef(int redefnum, char with_openssl, char binary_release,
-                  const char *package, const char *version);
+                  const char *package, const char *version, char netshare400);
 
 
 
@@ -127,7 +129,12 @@ int main(unsigned argc, char **argv) {
    char package[SMALLBUFSIZE+1], version[SMALLBUFSIZE+1];
    char debug[SMALLBUFSIZE+1];
    char with_openssl, binary_release='n';
-   int ch;
+   char ns400='n';
+   int ch, rc;
+
+if (argc>=2 && !strcmp(argv[1],"netshare400"))
+   ns400='y';
+
 
 /* Get the PACKAGE and VERSION info from configure.in */
 
@@ -151,7 +158,7 @@ int main(unsigned argc, char **argv) {
    strcpy(lib5250_objs, lib5250_sources);
    replaceall(lib5250_objs, ".c", ".o");
 
-
+ 
 /* find out about making a binary release */
 
    do {
@@ -199,11 +206,14 @@ int main(unsigned argc, char **argv) {
 
   /* Make a tn5250-configure.h file */
 
+   printf("\nCreating tn5250-config.h...\n");
+
    if (create_tn5250_config_h(TN5250_CONFIG_H_IN, "tn5250-config.h", 
-           with_openssl, binary_release, package, version) < 0) {
+           with_openssl, binary_release, package, version, ns400) < 0) {
         exit(1);
    }
 
+   printf("Creating config.h...\n");
    if (create_config_h("tn5250-config.h", "config.h", version) < 0) {
         exit(1);
    }
@@ -212,18 +222,31 @@ int main(unsigned argc, char **argv) {
 
    debug[0] = '\0';
 
+   printf("Creating Makefile...\n");
    if (create_makefile("Makefile.win.in", "Makefile", openssl_lib,
          openssl_include, innosetupdir, lib5250_objs, 
          installdir, package, version, debug)<0) {
          exit(1);
    }
 
-   if (create_makefile("tn5250_innosetup.iss.in", "tn5250_innosetup.iss", 
-         openssl_lib, openssl_include, innosetupdir, lib5250_objs, 
-         installdir, package, version, debug)<0) {
-         exit(1);
+   if (ns400=='n') {
+      printf("Creating Win32 script for InnoSetup 2...\n");
+      if (create_makefile("tn5250_innosetup.iss.in", "tn5250_innosetup.iss", 
+            openssl_lib, openssl_include, innosetupdir,
+            lib5250_objs, installdir, package, version, debug)<0) {
+            exit(1);
+      }
+   }
+   else {
+      printf("Creating Netshare400 script for InnoSetup 2...\n");
+      if (create_makefile("tn5250_innosetup.iss.in", "tn5250_innosetup.iss", 
+            openssl_lib, openssl_include, innosetupdir,
+            lib5250_objs, installdir, package, version, debug)<0) {
+            exit(1);
+      }
    }
 
+   printf("Creating ..\\Makefile...\n");
    make_root_makefile("..\\Makefile");
 
    return 0;
@@ -521,6 +544,7 @@ void get_path(const char *prompt, const char *file, char *path, int maxlen) {
  *    char             binary_release -
  *    const char  *    package        -
  *    const char  *    version        -
+ *    char             netshare400    -
  * DESCRIPTION
  *    Create a create a config.h file.  This is really a quick,
  *    minimal routine to allow us to get by without having autoconf
@@ -529,7 +553,7 @@ void get_path(const char *prompt, const char *file, char *path, int maxlen) {
  *****/
 int create_tn5250_config_h(const char *infile, const char *outfile, 
                char with_openssl, char binary_release, const char *package, 
-               const char *version) {
+               const char *version, char netshare400) {
 
       FILE *in;
       FILE *out;
@@ -570,7 +594,7 @@ int create_tn5250_config_h(const char *infile, const char *outfile,
                     while (mredefs[i].from!=NULL) {
                        if (strstr(rec, mredefs[i].from)!=NULL) {
                            s = manual_redef(mredefs[i].def, with_openssl, 
-                                 binary_release, package, version);
+                                 binary_release, package, version, netshare400);
                            if (s[0]!='\0') {
                                changed = 1;
                                fprintf(out, "%s\n", s);
@@ -607,7 +631,7 @@ int create_tn5250_config_h(const char *infile, const char *outfile,
            i = 0;
            while (mredefs[i].from!=NULL) {
                 s = manual_redef(mredefs[i].def, with_openssl, 
-                           binary_release, package, version);
+                           binary_release, package, version, netshare400);
                 if (s[0]!='\0') {
                       fprintf(out, "%s\n", s);
                 }
@@ -829,6 +853,7 @@ int make_root_makefile(const char *fn) {
  *    int                binary_release -
  *    const char    *    package        -
  *    const char    *    version        -
+ *    char               netshare400    -
  * DESCRIPTION
  *    Certain #define's that need to be put in config.h
  *    require some calculations to determine what the macro
@@ -836,15 +861,17 @@ int make_root_makefile(const char *fn) {
  *    and returns the string to put in the config.h file.
  *****/
 char *manual_redef(int redefnum, char with_openssl, char binary_release,
-                   const char *package, const char *version) {
+                   const char *package, const char *version, char netshare400) {
 
      switch (redefnum) {
        case DEF_HAVE_LIBCRYPTO:
+         *my_redef='\0';
          if (with_openssl=='y') {
               strcpy(my_redef, "#define HAVE_LIBCRYPTO 1");
          }
          break;
        case DEF_HAVE_LIBSSL:
+         *my_redef='\0';
          if (with_openssl=='y') {
               strcpy(my_redef, "#define HAVE_LIBSSL 1");
          }
@@ -865,11 +892,21 @@ char *manual_redef(int redefnum, char with_openssl, char binary_release,
          sprintf(my_redef,"#define VERSION \"%s\"\n", version);
          break;
        case DEF_BINARY:
+         *my_redef = '\0';
          if (binary_release=='y') {
               strcpy(my_redef, "#define BINARY_RELEASE 1");
          }
          break;
+       case DEF_NETSHARE400:
+         *my_redef = '\0';
+         if (netshare400=='y') {
+              strcpy(my_redef, "#define NETSHARE400 1");
+         }
+         break;
      }
 
      return my_redef;
 }
+
+
+
